///////////////////////////////////////////////////////////////////////////
// Copyright (c) 2013 Academy of Motion Picture Arts and Sciences 
// ("A.M.P.A.S."). Portions contributed by others as indicated.
// All rights reserved.
// 
// A worldwide, royalty-free, non-exclusive right to copy, modify, create
// derivatives, and use, in source and binary forms, is hereby granted, 
// subject to acceptance of this license. Performance of any of the 
// aforementioned acts indicates acceptance to be bound by the following 
// terms and conditions:
//
//  * Copies of source code, in whole or in part, must retain the 
//    above copyright notice, this list of conditions and the 
//    Disclaimer of Warranty.
//
//  * Use in binary form must retain the above copyright notice, 
//    this list of conditions and the Disclaimer of Warranty in the
//    documentation and/or other materials provided with the distribution.
//
//  * Nothing in this license shall be deemed to grant any rights to 
//    trademarks, copyrights, patents, trade secrets or any other 
//    intellectual property of A.M.P.A.S. or any contributors, except 
//    as expressly stated herein.
//
//  * Neither the name "A.M.P.A.S." nor the name of any other 
//    contributors to this software may be used to endorse or promote 
//    products derivative of or based on this software without express 
//    prior written permission of A.M.P.A.S. or the contributors, as 
//    appropriate.
// 
// This license shall be construed pursuant to the laws of the State of 
// California, and any disputes related thereto shall be subject to the 
// jurisdiction of the courts therein.
//
// Disclaimer of Warranty: THIS SOFTWARE IS PROVIDED BY A.M.P.A.S. AND 
// CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
// BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS 
// FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO 
// EVENT SHALL A.M.P.A.S., OR ANY CONTRIBUTORS OR DISTRIBUTORS, BE LIABLE 
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, RESITUTIONARY, 
// OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
// THE POSSIBILITY OF SUCH DAMAGE.
//
// WITHOUT LIMITING THE GENERALITY OF THE FOREGOING, THE ACADEMY 
// SPECIFICALLY DISCLAIMS ANY REPRESENTATIONS OR WARRANTIES WHATSOEVER 
// RELATED TO PATENT OR OTHER INTELLECTUAL PROPERTY RIGHTS IN THE ACADEMY 
// COLOR ENCODING SYSTEM, OR APPLICATIONS THEREOF, HELD BY PARTIES OTHER 
// THAN A.M.P.A.S., WHETHER DISCLOSED OR UNDISCLOSED.
///////////////////////////////////////////////////////////////////////////

#include <string>
#include <fstream>
#include <iostream>
#include <assert.h>
#include <CtlSimdInterpreter.h>

using namespace std;
using namespace Ctl;

void
makeHugeArray(const string & moduleName)
{
    string filename = "./" + moduleName + ".ctl";
    
    cout << "\tGenerating file " << filename << endl;

    ofstream out(filename.c_str());
    assert(out);


    int size = int(1e6);
    
    out << "namespace testHugeInit\n{\n\n";
    out << "const int vlaSize = " << size << ";\n\n";
    out << "const int vla[] = {\n";

    for(int i = 0; i < size; i++)
    {
	out << i;
	if( i != size-1)
	    out << ", ";
	if( i%10 == 0)
	    out << endl;
    }
    
    out << "};\n";

    out << "const int vla2DSize = " << size << ";\n\n";
    out << "const int vla2D[][] = {{\n";

    for(int i = 0; i < size; i++)
    {
	out << i;
	if( i != size-1)
	    out << ", ";
	if( i%10 == 0)
	    out << endl;
    }
    out << "},{\n";
    
    for(int i = 0; i < size; i++)
    {
	out << -i;
	if( i != size-1)
	    out << ", ";
	if( i%10 == 0)
	    out << endl;
    }
    out << "}};\n";

    out << "} // namespace \"testHugeInit\"\n";
    cout << "\tok\n";
}

void
cleanupHugeArray(const string & moduleName)
{
    string filename = "./" + moduleName + ".ctl";
    
    cout << "\tRemoving file " << filename << endl;

    remove(filename.c_str());
    cout << "\tok\n";
}

void
testHugeInit()
{
    cout << "Testing a very large static array." << endl;
    try
    {
	string modulename = "ctlMakeBigArray";
	makeHugeArray(modulename);

	SimdInterpreter interp;

	cout << "\tLoading large arrays in module " << modulename << ".\n";
	interp.loadModule (modulename);
	interp.loadModule ("testHugeInit");

	cleanupHugeArray(modulename);

	cout << "\tok\n";
    }
    catch (const std::exception &e)
    {
	cerr << "ERROR -- caught exception: " << e.what() << endl;
	assert (false);
    }
    cout << "ok" << endl;

}
